from keyword import iskeyword
from typing import Literal, Optional

from pydantic import ConfigDict, Field, model_serializer, model_validator

from harp.services.models import ServiceDefinition

from .base import BaseConfigurable


class Service(BaseConfigurable):
    """A settings base class for service definitions."""

    model_config = ConfigDict(extra="allow")

    #: Base type for service definition. This is not usually the base interface that the service implements, and you
    #: should use the `type` field to override the actually instanciated type.
    base: Optional[str] = Field(default=None, description="Base type for service definition.")

    #: Type for service definition. This is the actual type that will be instanciated.
    type: Optional[str] = Field(default=None, description="Type for service definition.")

    #: Constructor for service definition. If provided, will be used instead of the default constructor.
    constructor: Optional[str] = Field(default=None, description="Optional custom constructor for the service.")

    #: Arguments for the service constructor, by name.
    arguments: Optional[dict] = Field(default_factory=dict, description="Arguments for the service constructor.")

    @model_validator(mode="before")
    @classmethod
    def __validate(cls, values):
        """Use extra fields as arguments."""
        arguments = values.get("arguments", {})
        for k in list(values.keys()):
            if k not in cls.model_fields:
                if str.isidentifier(k) and not iskeyword(k):
                    arguments[k] = values.pop(k)
                else:
                    raise ValueError(f"Invalid field name: {k} for {cls.__name__}")
        if len(arguments):
            return {**values, "arguments": arguments}
        return values

    @model_serializer(mode="wrap")
    def __serialize(self, wrapped, context):
        """Enhance serialization logic to inline arguments, unless they conflict with a model field."""
        data = wrapped(self, context)
        arguments = data.pop("arguments", {})
        inline_arguments = {}
        for k in list(arguments.keys()):
            if k not in type(self).model_fields:
                inline_arguments[k] = arguments.pop(k)

        for k in ("base", "type", "constructor"):
            if k in data and data[k] is None:
                data.pop(k)

        return {
            **data,
            **({"arguments": arguments} if len(arguments) else {}),
            **inline_arguments,
        }

    def to_service_definition(
        self, name: str, lifestyle: Optional[Literal["singleton", "transient", "scoped"]] = "singleton"
    ) -> ServiceDefinition:
        """Convert the service settings to a service definition."""
        return ServiceDefinition(
            name=name,
            base=self.base,
            type=self.type,
            constructor=self.constructor,
            arguments=self.arguments,
            lifestyle=lifestyle,
        )


class LazyService(BaseConfigurable):
    """A lazy service definition, that will be resolved at runtime."""

    type: str | list[str] = Field(..., description="Reference to the service to resolve at runtime.")

    def resolve(self, resolver, context):
        """
        Resolve reference value in using the given resolver (callable) and resolution context.
        """
        return resolver(self.type, context)

    @model_serializer(mode="wrap")
    def __serialize(self, wrapped, context):
        """Enhance serialization logic to inline arguments, unless they conflict with a model field."""
        return wrapped(self, context)



from typing import Self, Type

from pydantic import BaseModel, ConfigDict


class BaseConfigurable(BaseModel):
    """
    Base class for configurables (a.k.a settings definitions). For your settings definitions, prefer
    :class:`Configurable` over this base class (see below).

    Extends Pydantic's BaseModel.

    """

    @classmethod
    def from_dict(cls: Type[Self], data: dict) -> Self:
        """Construct a new instance from a dictionary."""
        return cls(**data)

    @classmethod
    def from_kwargs(cls: Type[Self], **kwargs) -> Self:
        """Construct a new instance from keyword arguments."""
        return cls(**kwargs)

    def __getitem__(self, item):
        """Allow access to nested values using dot notation."""
        value = self
        for bit in item.split("."):
            value = getattr(value, bit)
        return value


class Configurable(BaseConfigurable):
    """A settings definition base."""

    model_config = ConfigDict(extra="forbid")



"""
Dashboard Application

"""

from os.path import dirname
from pathlib import Path

from harp import get_logger
from harp.config import Application, OnBindEvent, OnBoundEvent

from .settings import DashboardSettings

logger = get_logger(__name__)


async def on_bind(event: OnBindEvent):
    # Load service definitions, bound to our settings.
    event.container.load(
        Path(dirname(__file__)) / "services.yml",
        bind_settings=event.settings["dashboard"],
    )


async def on_bound(event: OnBoundEvent):
    # This should be feasible using the declarative approach, but for now, we need to ensure the instances are created
    # and alive for them to be registered with the router. The control order is probably wrong, it would be better if
    # the children controllers were injected as a list to the parent controller, but RoutingControllers take the router
    # instance, for now, and register themselves.
    event.provider.set(
        "dashboard.subcontrollers",
        [
            event.provider.get("dashboard.controller.blobs"),
            event.provider.get("dashboard.controller.overview"),
            event.provider.get("dashboard.controller.system"),
            event.provider.get("dashboard.controller.transactions"),
        ],
    )

    # Add our controller to the controller resolver, using the configured dashboard port. This is what will actually
    # make the server to route requests to the dashboard controller when an incoming request is received on the
    # dashboard port.
    event.resolver.add_controller(
        event.provider.get("dashboard.controller"),
        name="dashboard",
        port=event.provider.get(DashboardSettings).port,
    )


application = Application(
    settings_type=DashboardSettings,
    on_bind=on_bind,
    on_bound=on_bound,
    dependencies=["storage"],
)



from datetime import UTC, datetime, timedelta
from decimal import Decimal
from statistics import StatisticsError, mean

from harp import get_logger
from harp.controllers import GetHandler, RouterPrefix, RoutingController
from harp.http import HttpRequest
from harp.views import json
from harp_apps.storage.constants import TimeBucket
from harp_apps.storage.types import IStorage

from ..utils.dates import generate_continuous_time_range, get_start_datetime_from_range

logger = get_logger(__name__)

time_bucket_for_range = {
    "1h": TimeBucket.HOUR.value,
    "24h": TimeBucket.HOUR.value,
    "7d": TimeBucket.DAY.value,
    "1m": TimeBucket.DAY.value,
    "1y": TimeBucket.DAY.value,
}


def _format_aggregate(items, count, key, *, default=None):
    if count >= 86400:
        period = "sec"
        divider = 86400

    elif count >= 1440:
        period = "min"
        divider = 1440

    elif count >= 24:
        period = "hour"
        divider = 24

    else:
        period = "day"
        divider = 1

    return {
        "rate": Decimal(int(10 * (count / divider))) / 10,
        "period": period,
        "data": [
            {
                "datetime": t["datetime"],
                "value": int(t[key]) if t[key] is not None else default,
            }
            for t in items
        ],
    }


@RouterPrefix("/api/overview")
class OverviewController(RoutingController):
    def __init__(self, *, storage: IStorage, handle_errors=True, router=None):
        self.storage = storage
        super().__init__(handle_errors=handle_errors, router=router)

    @GetHandler("/summary")
    async def get_summary_data(self, request: HttpRequest):
        time_span = "24h"
        time_bucket = time_bucket_for_range[time_span]
        now = datetime.now(UTC)
        start_datetime = get_start_datetime_from_range(time_span, now=now + timedelta(hours=1))

        transactions_by_date_list = await self.storage.transactions_grouped_by_time_bucket(
            start_datetime=start_datetime, time_bucket=time_bucket
        )

        errors_count = sum([t["errors"] for t in transactions_by_date_list])
        transactions_count = sum([t["count"] for t in transactions_by_date_list])

        transactions_by_date_list = generate_continuous_time_range(
            discontinuous_transactions=transactions_by_date_list,
            time_bucket=time_bucket,
            start_datetime=start_datetime,
        )
        mean_tpdex = _calculate_mean_tpdex(transactions_by_date_list)

        return json(
            {
                "tpdex": {
                    "mean": int(mean_tpdex),
                    "data": [
                        {
                            "datetime": t["datetime"],
                            "value": (int(t["meanTpdex"]) if t["meanTpdex"] is not None else 100),
                        }
                        for t in transactions_by_date_list
                    ],
                },
                "transactions": _format_aggregate(transactions_by_date_list, transactions_count, "count", default=0),
                "errors": _format_aggregate(transactions_by_date_list, errors_count, "errors", default=0),
            }
        )

    @GetHandler("/")
    async def get_overview_data(self, request: HttpRequest):
        # endpoint and range from request
        endpoint = request.query.get("endpoint")
        range = request.query.get("timeRange", "24h")

        # time buckets and start_datetime accordingly
        time_bucket = time_bucket_for_range.get(range, "day")
        start_datetime = get_start_datetime_from_range(range)

        logger.debug(
            f"🛑 {type(self).__name__}::get_overview_data 1️⃣ ",
            endpoint=endpoint,
            range=range,
            time_bucket=time_bucket,
            start_datetime=start_datetime,
        )

        transactions_by_date_list = await self.storage.transactions_grouped_by_time_bucket(
            endpoint=endpoint,
            start_datetime=start_datetime,
            time_bucket=time_bucket,
        )

        logger.debug(
            f"🛑 {type(self).__name__}::get_overview_data 2️⃣ ",
            transactions_by_date_list=transactions_by_date_list,
        )

        errors_count = sum([t["errors"] for t in transactions_by_date_list])
        transactions_count = sum([t["count"] for t in transactions_by_date_list])
        errors_rate = errors_count / transactions_count if transactions_count else 0
        mean_duration = (
            sum([t["meanDuration"] * t["count"] for t in transactions_by_date_list]) / transactions_count
            if transactions_count
            else 0
        )
        transactions_by_date_list = generate_continuous_time_range(
            discontinuous_transactions=transactions_by_date_list,
            time_bucket=time_bucket,
            start_datetime=start_datetime,
        )

        logger.debug(
            f"🛑 {type(self).__name__}::get_overview_data 3️⃣ ",
            transactions_by_date_list=transactions_by_date_list,
        )

        mean_tpdex = _calculate_mean_tpdex(transactions_by_date_list)

        result = {
            "transactions": transactions_by_date_list,
            "errors": {"count": errors_count, "rate": errors_rate},
            "count": transactions_count,
            "meanDuration": mean_duration,
            "meanTpdex": mean_tpdex,
            "timeRange": range,
        }

        logger.debug(
            f"🛑 {type(self).__name__}::get_overview_data 4️⃣ ",
            result=result,
        )

        return json(result)


def _calculate_mean_tpdex(transactions_by_date_list):
    try:
        return mean(filter(None, [t["meanTpdex"] for t in transactions_by_date_list]))
    except StatisticsError:
        return 100



import hashlib

import pytest

from harp.utils.collections import all_combinations
from harp.utils.testing.http import parametrize_with_http_methods


@pytest.mark.parametrize(
    "parameters",
    list(
        sorted(
            all_combinations(
                [
                    "include_safe",
                    "include_unsafe",
                    "include_idempotent",
                    "include_non_idempotent",
                    "include_standard",
                    "include_non_standard",
                    "include_having_request_body",
                    "include_maybe_having_request_body",
                    "include_not_having_request_body",
                    "include_having_response_body",
                    "include_maybe_having_response_body",
                    "include_not_having_response_body",
                ],
                maxlen=3,
            ),
        )
    ),
    ids=lambda x: hashlib.md5(",".join(sorted(x)).encode()).hexdigest()[:7],
)
def test_methods(snapshot, parameters):
    p = parametrize_with_http_methods(**{k: True for k in parameters})
    assert snapshot == ",".join(sorted(parameters)) + " => " + ",".join(sorted(p.args[1]))



from re import escape
from unittest.mock import ANY

import orjson

from harp.utils.testing import RE
from harp.utils.testing.communicators import ASGICommunicator
from harp.utils.testing.databases import parametrize_with_blob_storages_urls, parametrize_with_database_urls
from harp.utils.testing.mixins import ControllerThroughASGIFixtureMixin
from harp_apps.dashboard.controllers.system import SystemController
from harp_apps.dashboard.tests._mixins import SystemControllerTestFixtureMixin, parametrize_with_settings
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class TestSystemController(
    SystemControllerTestFixtureMixin,
    StorageTestFixtureMixin,
):
    @parametrize_with_settings({})
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("sql")
    async def test_get_settings_empty(self, controller: SystemController):
        response = await controller.get_settings()
        assert response.asdict() == {
            "applications": ["harp_apps.storage"],
            "storage": {
                "blobs": {"type": "sql"},
                "migrate": True,
                "url": "sqlite+aiosqlite:///:memory:",
                "redis": None,
            },
        }

    @parametrize_with_settings({"storage": {"redis": {"url": "redis://redis.example.com:1234/42"}}})
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("redis")
    async def test_get_settings_redis(self, controller: SystemController):
        response = await controller.get_settings()
        assert response.asdict() == {
            "applications": ["harp_apps.storage"],
            "storage": {
                "blobs": {"type": "redis"},
                "migrate": True,
                "url": "sqlite+aiosqlite:///:memory:",
                "redis": {"url": "redis://redis.example.com:1234/42"},
            },
        }

    @parametrize_with_settings({"storage": {"url": "sqlite+aiosqlite:///:memory:"}})
    @parametrize_with_database_urls("sqlite")
    async def test_get_settings_sqlalchemy(self, controller: SystemController):
        response = await controller.get_settings()
        assert response.asdict() == {
            "applications": ["harp_apps.storage"],
            "storage": {
                "blobs": ANY,
                "migrate": True,
                "url": "sqlite+aiosqlite:///:memory:",
                "redis": None,
            },
        }

    @parametrize_with_settings({"applications": ["storage"]})
    @parametrize_with_database_urls("postgresql")
    async def test_get_settings_sqlalchemy_secure(self, controller: SystemController):
        response = await controller.get_settings()
        assert response.asdict() == {
            "applications": ["harp_apps.storage"],
            "storage": {
                "migrate": True,
                "url": RE(r".*://test:\*\*\*@.*"),
                "blobs": ANY,
                "redis": None,
            },
        }


class TestSystemControllerThroughASGI(
    SystemControllerTestFixtureMixin,
    StorageTestFixtureMixin,
    ControllerThroughASGIFixtureMixin,
):
    @parametrize_with_settings({})
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("sql")
    async def test_get_settings_empty(self, client: ASGICommunicator):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert response["body"] == (
            b'{"applications":["harp_apps.storage"],"storage":{"url":"sqlite+aiosqlite:///'
            b':memory:","migrate":true,"blobs":{"type":"sql"},"redis":null}}'
        )

    @parametrize_with_settings({})
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("redis")
    async def test_get_settings_empty_redis(self, client: ASGICommunicator):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert response["body"] == (
            b'{"applications":["harp_apps.storage"],"storage":{"url":"sqlite+aiosqlite:///'
            b':memory:","migrate":true,"blobs":{"type":"redis"},"redis":null}}'
        )

    @parametrize_with_settings(
        {
            "applications": ["storage"],
            "storage": {"url": "sqlite+aiosqlite:///:memory:"},
        }
    )
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("sql")
    async def test_get_settings_sqlalchemy(self, client: ASGICommunicator, blob_storage):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert response["body"] == (
            b'{"applications":["harp_apps.storage"],"storage":{"url":"sqlite+aiosqlite:///'
            b':memory:","migrate":true,"blobs":{"type":"sql"},"redis":null}}'
        )

    @parametrize_with_settings(
        {
            "applications": ["storage"],
            "storage": {
                "url": "sqlite+aiosqlite:///:memory:",
                "blobs": {"type": "redis"},
            },
        }
    )
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("redis")
    async def test_get_settings_sqlalchemy_blobs_in_redis(self, client: ASGICommunicator):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert response["body"] == (
            b'{"applications":["harp_apps.storage"],"storage":{"url":"sqlite+aiosqlite:///'
            b':memory:","migrate":true,"blobs":{"type":"redis"},"redis":null}}'
        )

    @parametrize_with_database_urls("postgresql")
    @parametrize_with_blob_storages_urls("sql")
    async def test_get_settings_sqlalchemy_secure(self, client: ASGICommunicator, blob_storage):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert orjson.loads(response["body"]) == {
            "applications": ["harp_apps.storage"],
            "storage": {
                "blobs": {"type": "sql"},
                "migrate": True,
                "redis": None,
                "url": RE(escape("postgresql+asyncpg://test:***@") + r".*"),
            },
        }

    @parametrize_with_settings(
        {
            "applications": ["storage"],
            "storage": {
                "blobs": {"type": "redis"},
                "redis": {"url": "redis://user:password@localhost:6379/0"},
            },
        }
    )
    @parametrize_with_database_urls("sqlite")
    @parametrize_with_blob_storages_urls("redis")
    async def test_get_settings_sqlalchemy_blobs_in_redis_secure(self, client: ASGICommunicator, blob_storage):
        response = await client.http_get("/api/system/settings")

        assert response["status"] == 200
        assert response["headers"] == ((b"content-type", b"application/json"),)
        assert response["body"] == (
            b'{"applications":["harp_apps.storage"],"storage":{"url":"sqlite+aiosqlite:///'
            b':memory:","migrate":true,"blobs":{"type":"redis"},"redis":{"url":"redis://us'
            b'er:***@localhost:6379/0"}}}'
        )



import pytest

from harp.utils.testing.communicators import ASGICommunicator
from harp.utils.testing.mixins import ControllerThroughASGIFixtureMixin
from harp_apps.dashboard.controllers.blobs import BlobsController
from harp_apps.storage.models import Blob
from harp_apps.storage.types import IBlobStorage
from harp_apps.storage.utils.testing.mixins import StorageTestFixtureMixin


class BlobsControllerTestFixtureMixin:
    @pytest.fixture
    async def controller(self, blob_storage):
        return BlobsController(storage=blob_storage, handle_errors=False)


class TestBlobsController(
    BlobsControllerTestFixtureMixin,
    StorageTestFixtureMixin,
):
    async def test_get_not_found(self, controller):
        response = await controller.get("not-a-blob")
        assert response.status == 404
        assert response.content_type == "text/plain"

    async def test_get_existing(self, controller, blob_storage: IBlobStorage):
        await blob_storage.put(Blob(id="blob-1", data=b"hello"))

        response = await controller.get("blob-1")
        assert response.status == 200
        assert response.content_type == "application/octet-stream"

        response = await controller.get("blob-2")
        assert response.status == 404
        assert response.content_type == "text/plain"


class TestBlobsControllerThroughASGI(
    BlobsControllerTestFixtureMixin,
    StorageTestFixtureMixin,
    ControllerThroughASGIFixtureMixin,
):
    async def test_get_not_found(self, client: ASGICommunicator):
        response = await client.http_get("/api/blobs/not-a-blob")

        assert response["status"] == 404
        assert response["headers"] == ((b"content-type", b"text/plain"),)



from tempfile import NamedTemporaryFile

from sqlalchemy import make_url

from harp.config import ConfigurationBuilder
from harp.config.asdict import asdict


def create_settings(settings=None):
    config = ConfigurationBuilder(settings)
    config.applications.add("storage")
    config.applications.add("dashboard")
    return asdict(config.build())


class TestSettings:
    async def test_dashboard_auth_basic(self):
        users = {"foo": "bar"}
        settings = create_settings({"dashboard": {"auth": {"type": "basic", "algorithm": "plain", "users": users}}})
        assert settings["dashboard"]["auth"] == {
            "algorithm": "plaintext",
            "type": "basic",
            "users": {"foo": {"password": "bar"}},
        }

    async def test_dashboard_auth_basic_from_file(self):
        with NamedTemporaryFile("w", delete_on_close=False) as f:
            f.write("romain: s3cr3t")
            f.close()
            settings = create_settings({"dashboard": {"auth": {"type": "basic", "users": {"fromFile": f.name}}}})

            assert settings["dashboard"]["auth"] == {
                "type": "basic",
                "users": {"romain": {"password": "s3cr3t"}},
            }

    async def test_sqlalchemy_url_merges(self):
        dburl = "postgresql+asyncpg://user:pass@localhost:1234/name"

        builder = ConfigurationBuilder()
        builder.applications.add("storage")
        builder.add_values({"storage": {"url": dburl}})
        builder.add_values({"storage": {"url": dburl}})
        config = builder.build()
        assert str(config["storage"].url) == make_url(dburl).render_as_string(hide_password=False)
        assert asdict(config, secure=False)["storage"]["url"] == dburl



from pydantic import BaseModel


class GlobalSettings(dict[str, BaseModel | list]):
    """
    Placeholder type for injecting global settings as a dictionary. Usually, you'd prefer to use the "local" settings
    of your application, but sometimes you need to access the global settings. For example, the dashboard uses it to
    display all the running instance settings in a system tab.
    """



import asyncio

import orjson
from rich.console import Console
from rich.pretty import Pretty
from rich.syntax import Syntax
from rich.tree import Tree

from harp.commandline.options.server import CommonServerOptions, server_command
from harp.config import ConfigurationBuilder
from harp.config.asdict import asdict
from harp.utils.commandline import click


@server_command("config", short_help="Prints the current configuration.")
@click.option("--raw", is_flag=True, help="Prints the raw configuration as a dictionary.")
@click.option("--json", is_flag=True, help="Prints the raw configuration as JSON.")
@click.option(
    "--unsecure",
    is_flag=True,
    help="Prints the configuration without hiding sensitive information.",
)
def config(raw=False, json=False, unsecure=False, **kwargs):
    """Compiles and dumps the current configuration, the same way that "server" commands would do it.

    Example:

        $ harp config --file ... --example ... --set ... --endpoint ...

    """
    if raw and json:
        raise click.UsageError("Cannot use both --raw and --json.")

    system = asyncio.run(
        ConfigurationBuilder.from_commandline_options(
            CommonServerOptions(**kwargs),
        ).abuild_system()
    )

    console = Console()
    if raw:
        console.print(Pretty(asdict(system.config)))
    elif json:
        console.print(
            Syntax(
                orjson.dumps(
                    asdict(
                        system.config,
                        secure=not unsecure,
                    ),
                    option=orjson.OPT_INDENT_2,
                ).decode(),
                "json",
                background_color="default",
            )
        )
    else:
        for k, v in system.config.items():
            tree = Tree(f"📦 [bright_white][bold]{k}[/bright_white][/bold]")
            # todo secure/unsecure with pretty print ?
            tree.add(Pretty(v))
            console.print(tree)



from harp.config.asdict import asdict

from ..settings import AcmeSettings


def test_default_settings():
    settings = AcmeSettings()
    assert asdict(settings) == {}
    assert asdict(settings, verbose=True) == {"owner": "Joe"}


def test_custom_settings():
    settings = AcmeSettings(owner="Alice")
    assert asdict(settings) == {"owner": "Alice"}
    assert asdict(settings, verbose=True) == {"owner": "Alice"}



"""
The Settings (:mod:`harp.settings`) module contains hardcoded configuration for Harp. This is probably only handy
for the young days of harp, and will be replaced by something less hard-coded later.

Contents
--------

"""

__title__ = "Settings"

import builtins
from os import environ

#: Pagination size for api endpoints
PAGE_SIZE = 40

#: Default timeout for http requests
DEFAULT_TIMEOUT = 30.0

#: Force environment override (dev or prod)
HARP_ENV = environ.pop("HARP_ENV", None)
if HARP_ENV is not None:
    HARP_ENV = HARP_ENV.strip().lower()
    if HARP_ENV not in ("dev", "prod"):
        HARP_ENV = None


def is_test_context():
    return getattr(builtins, "__pytest__", False)



from typing import Optional

from pydantic import Field

from harp.config import Configurable


class DevserverSettings(Configurable):
    enabled: bool = Field(
        True,
        description="Enable or disable the development server.",
    )
    port: Optional[int] = Field(
        None,
        description=(
            "Port on which the development server will be served (internal). The proxy will forward dashboard "
            "requests to this port, if enabled."
        ),
    )



"""rename apdex to tpdex

Revision ID: 0b4d9cb71c38
Revises: 248285939c4c
Create Date: 2024-06-14 15:28:02.344136

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "0b4d9cb71c38"
down_revision: Union[str, None] = "248285939c4c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.alter_column(
        "transactions",
        "apdex",
        new_column_name="tpdex",
        existing_type=sa.Integer(),
        nullable=True,
    )


def downgrade() -> None:
    op.alter_column(
        "transactions",
        "tpdex",
        new_column_name="apdex",
        existing_type=sa.Integer(),
        nullable=True,
    )



from typing import Annotated, Any, Iterable, Literal, Mapping, Optional, Self, Sequence, Tuple, Union

from pydantic import BaseModel, ConfigDict, Discriminator, Tag

from harp.utils.config import yaml

from .references import LazyServiceReference, LazySettingReference

StringOrRef = str | LazySettingReference | LazyServiceReference
ExtendedStringOrRef = StringOrRef | bool


def _resolve(value: Optional[ExtendedStringOrRef | Iterable[ExtendedStringOrRef]], settings: Any):
    if value is None:
        return None
    if isinstance(value, (bool, str, LazyServiceReference)):
        return value
    if isinstance(value, Mapping):
        return {k: _resolve(v, settings) for k, v in value.items()}
    if isinstance(value, LazySettingReference):
        return value.resolve(settings)
    for x in value:
        x = _resolve(x, settings)
        if x is not None:
            return x
    return None


class ServiceDefinition(BaseModel):
    """
    Describes a service that our container is able to register. Services within a collection can override each other,
    if explicitely stated. This is useful for conditional service change (for example, based on a configuration value).

    """

    model_config = ConfigDict(extra="forbid")

    #: service name (todo: constraints ?)
    name: str

    #: service description, for documentation and readbility purposes
    description: Optional[str] = None

    #: service lifestyle
    lifestyle: Optional[Literal["singleton", "transient", "scoped"]] = None

    #: condition to be met for the service to register or override another one
    override: Optional[str] = None

    #: base type for service, aka the interface we are implementing, if different from type
    base: Optional[StringOrRef | Sequence[StringOrRef]] = None

    #: service type, aka the implementation we are using, if different from base
    type: Optional[StringOrRef | Sequence[StringOrRef]] = None

    #: constructor name, if not default one
    constructor: Optional[StringOrRef | Sequence[StringOrRef]] = None

    #: default named arguments for the service constructor. Can be both positionnal or keyword based, but the
    # parameter choice is always based on parameter name, by key.
    defaults: Optional[Mapping[str, Any] | Sequence[Mapping[str, Any] | LazySettingReference]] = None

    #: named arguments for the service constructor. Mapped after the annotation parsing, so it can override default
    # attributions from type annotations.
    arguments: Optional[Mapping[str, Any] | Sequence[Mapping[str, Any] | LazySettingReference]] = None

    #: positionnal arguments for the service constructor.
    positionals: Optional[Tuple[Any, ...]] = None

    def override_with(self, other: Self) -> Self:
        if self.name != other.name:
            raise ValueError(f"Service name mismatch: {self.name} != {other.name}")

        # todo scope override limitation, as it is most probably a mistake.

        return type(self)(
            name=self.name,
            description=(other.description if other.description is not None else self.description),
            lifestyle=(other.lifestyle if other.lifestyle is not None else self.lifestyle),
            base=other.base if other.base is not None else self.base,
            type=other.type if other.type is not None else self.type,
            constructor=(other.constructor if other.constructor is not None else self.constructor),
            arguments=(other.arguments if other.arguments is not None else self.arguments),
            defaults=other.defaults if other.defaults is not None else self.defaults,
            positionals=(other.positionals if other.positionals is not None else self.positionals),
        )

    def bind_settings(self, settings: Any):
        for k in ("base", "type", "constructor"):
            setattr(self, k, _resolve(getattr(self, k), settings))

        self.defaults = _resolve(self.defaults, settings)
        self.arguments = _resolve(self.arguments, settings)


def _discriminator(obj):
    if "services" in obj:
        if "condition" in obj:
            return "conditional_collection"
        return "collection"
    return "service"


class BaseServiceDefinitionCollection(BaseModel):
    """
    Base class for coherent sequences of services. The traverse() method can be used to get a flat iterator over all
    the service definitions in the collection and its children, ordered.

    """

    services: Sequence[
        Annotated[
            Union[
                Annotated["ServiceDefinition", Tag("service")],
                Annotated["BaseServiceDefinitionCollection", Tag("collection")],
                Annotated["ConditionalServiceDefinitionCollection", Tag("conditional_collection")],
            ],
            Discriminator(_discriminator),
        ]
    ]

    def traverse(self) -> Iterable[ServiceDefinition]:
        for service_or_collection in iter(self.services):
            if isinstance(service_or_collection, ServiceDefinition):
                yield service_or_collection
            else:
                yield from service_or_collection.traverse()

    def bind_settings(self, settings: Any):
        for service in self.services:
            service.bind_settings(settings)


class ConditionalServiceDefinitionCollection(BaseServiceDefinitionCollection):
    """
    A collection of services that are only registered if a condition is met. This is useful for conditional service
    registration, for example based on a configuration value.

    """

    condition: Optional[
        Union[
            str | bool | LazySettingReference,
            Sequence[str | bool | LazySettingReference],
        ]
    ] = None

    def bind_settings(self, settings: Any):
        if self.condition:
            self.condition = _resolve(self.condition, settings)
        super().bind_settings(settings)

    def traverse(self) -> Iterable[ServiceDefinition]:
        if self.condition:
            yield from super().traverse()


class ServiceDefinitionCollection(BaseServiceDefinitionCollection):
    """
    Final class for a service collection. Iterate on it to get a flattened (one level) and merged (services with same
    name that allows overrides are merged together) list of services.

    """

    def __iter__(self) -> Iterable[ServiceDefinition]:
        _map = {}
        for service in self.traverse():
            if service.name in _map:
                if not service.override:
                    raise ValueError(
                        f"Service with name {service.name} is already defined, but no override flag is set."
                    )
                _map[service.name] = _map[service.name].override_with(service)
                continue

            if service.override:
                raise ValueError(f"Service with name {service.name} is not defined, but override flag is set.")
            _map[service.name] = service

        return iter(_map.values())

    @classmethod
    def model_validate_yaml(cls, filename):
        return cls.model_validate(yaml.load(filename))



from unittest.mock import AsyncMock

import respx
from httpx import AsyncClient, Response

from harp.config.asdict import asdict
from harp_apps.http_client.contrib.hishel.storages import AsyncStorage
from harp_apps.http_client.events import EVENT_FILTER_HTTP_CLIENT_REQUEST, EVENT_FILTER_HTTP_CLIENT_RESPONSE
from harp_apps.http_client.tests._base import BaseTestDefaultsWith
from harp_apps.storage.services.blob_storages.memory import MemoryBlobStorage

URL = "http://www.example.com/"


class TestDefaultsWithNoStorage(BaseTestDefaultsWith):
    async def test_defaults(self):
        system = await self.create_system()

        assert asdict(system.config) == {
            "applications": ["harp_apps.http_client"],
            "http_client": {},
        }
        assert asdict(system.config, verbose=True) == {
            "applications": ["harp_apps.http_client"],
            "http_client": {
                "cache": {
                    "controller": {
                        "allow_heuristics": False,
                        "allow_stale": False,
                        "cacheable_methods": ["GET", "HEAD"],
                        "cacheable_status_codes": [
                            200,
                            203,
                            204,
                            206,
                            300,
                            301,
                            308,
                            404,
                            405,
                            410,
                            414,
                            501,
                        ],
                        "type": "hishel.Controller",
                    },
                    "enabled": True,
                    "storage": {
                        "base": "hishel.AsyncBaseStorage",
                        "check_ttl_every": 60.0,
                        "ttl": None,
                        "type": "harp_apps.http_client.contrib.hishel.storages.AsyncStorage",
                    },
                    "transport": {"type": "hishel.AsyncCacheTransport"},
                },
                "proxy_transport": {"type": "harp_apps.http_client.transport.AsyncFilterableTransport"},
                "timeout": 30.0,
                "transport": {
                    "retries": 0,
                    "type": "httpx.AsyncHTTPTransport",
                    "verify": True,
                },
                "type": "httpx.AsyncClient",
            },
        }

        storage = system.provider.get("http_client.cache.storage")
        assert isinstance(storage, AsyncStorage)
        assert isinstance(storage._storage, MemoryBlobStorage)

    @respx.mock
    async def test_events(self):
        system = await self.create_system()
        endpoint = respx.get(URL).mock(return_value=Response(200, content=b"Hello, world."))
        http_client = system.provider.get(AsyncClient)

        # register events
        on_filter_request, on_filter_response = AsyncMock(), AsyncMock()
        system.dispatcher.add_listener(EVENT_FILTER_HTTP_CLIENT_REQUEST, on_filter_request)
        system.dispatcher.add_listener(EVENT_FILTER_HTTP_CLIENT_RESPONSE, on_filter_response)

        # simple request
        response = await http_client.get(URL)
        assert endpoint.called
        assert response.status_code == 200
        assert on_filter_request.called
        assert on_filter_response.called



import os


def cast_bool(x):
    if isinstance(x, str):
        value = x.lower()
        if value in ("true", "yes", "1"):
            return True
        elif value in ("false", "no", "0"):
            return False
        else:
            raise ValueError(f"Invalid string value: {x}")
    return bool(x)


def get_bool_from_env(key, default):
    value = os.environ.get(key, None)

    if value is None:
        return default

    try:
        return cast_bool(value)
    except ValueError:
        raise ValueError(f"Invalid boolean value for {key}: {value!r}")
